/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include "jabber/jabber.h"
//#include "ns.h" 



/* local macros for launching event handlers */
#define STATE_EVT(arg) if(j->on_state) { (j->on_state)(j, (arg) ); }

void JABBERCleanUp();

int SockBuffStandartSize = 8196;




/* prototypes of the local functions */
static void startElement(void *userdata, const char *name, const char **attribs);
static void endElement(void *userdata, const char *name);
static void charData(void *userdata, const char *s, int slen);

/*
 *  jab_new -- initialize a new jabber connection
 *
 *  parameters
 *      user -- jabber id of the user
 *      pass -- password of the user
 *
 *  results
 *      a pointer to the connection structure
 *      or NULL if allocations failed
 */
jconn jab_new(char *user, char *pass)
{
    pool p;
    jconn j;

    if(!user) return(NULL);

    p = pool_new();
    if(!p) return(NULL);
    j = pmalloc_x(p, sizeof(jconn_struct), 0);
    if(!j) return(NULL);
    j->p = p;

    j->user = jid_new(p, user);
    j->pass = pstrdup(p, pass);

    j->state = JCONN_STATE_OFF;
    j->id = 1;
    j->fd = -1;
    FD_ZERO( j->fdsetw = pmalloc(p , sizeof(fd_set)) );
    FD_ZERO( j->fdsetr = pmalloc(p , sizeof(fd_set)) );
    return j;
}

/*
 *  jab_delete -- free a jabber connection
 *
 *  parameters
 *      j -- connection
 *
 */
void jab_delete(jconn j)
{
    if(!j) return;

    jab_stop(j);
    pool_free(j->p);
}

/*
 *  jab_state_handler -- set callback handler for state change
 *
 *  parameters
 *      j -- connection
 *      h -- name of the handler function
 */
void jab_state_handler(jconn j, jconn_state_h h)
{
    if(!j) return;

    j->on_state = h;
}

/*
 *  jab_packet_handler -- set callback handler for incoming packets
 *
 *  parameters
 *      j -- connection
 *      h -- name of the handler function
 */
void jab_packet_handler(jconn j, jconn_packet_h h)
{
    if(!j) return;

    j->on_packet = h;
}


/*
 *  jab_start -- start connection
 *
 *  parameters
 *      j -- connection
 *
 */
void jab_start(jconn j)
{
    xmlnode x;
    char *t,*t2;

    if(!j || j->state != JCONN_STATE_OFF) return;

    j->parser = XML_ParserCreate(NULL);
    XML_SetUserData(j->parser, (void *)j);
    XML_SetElementHandler(j->parser, startElement, endElement);
    XML_SetCharacterDataHandler(j->parser, charData);

    j->fd = make_netsocket(5222, j->user->server, NETSOCKET_CLIENT);
    if(j->fd < 0) {
        //STATE_EVT(JCONN_STATE_OFF)
        return;
	}
    j->state = JCONN_STATE_CONNECTED;
    //STATE_EVT(JCONN_STATE_CONNECTED)

    /* start stream */
    x = jutil_header(NS_CLIENT, j->user->server);
    t = xmlnode2str(x);
    /* this is ugly, we can create the string here instead of jutil_header */
    /* what do you think about it? -madcat */
    t2 = strstr(t,"/>");
    *t2++ = '>';
    *t2 = '\0';
    jab_send_raw(j,"<?xml version='1.1'?>");
    jab_send_raw(j,t);
    xmlnode_free(x);

    j->state = JCONN_STATE_ON;
    //STATE_EVT(JCONN_STATE_ON)

}

/*
 *  jab_stop -- stop connection
 *
 *  parameters
 *      j -- connection
 */
void jab_stop(jconn j)
{   
    
   if(!j || j->state == JCONN_STATE_OFF) return;


   Ns_RWLockUnlock(j->lock);     //Lock for writing before we change the values
   Ns_RWLockWrLock(j->lock);    // of the jconn object
   
   if(j->state == JCONN_STATE_OFF) return; 
    j->state = JCONN_STATE_OFF;
    close(j->fd);
    j->fd = -1;
    if(j->parser) 
      { 
       XML_ParserFree(j->parser); 
       j->parser = NULL;
      }
   JABBERCleanUp();              //Clean up the online data in the db
   Ns_RWLockUnlock(j->lock);
   Ns_RWLockRdLock(j->lock);
}

/*
 *  jab_getfd -- get file descriptor of connection socket
 *
 *  parameters
 *      j -- connection
 *
 *  returns
 *      fd of the socket or -1 if socket was not connected
 */
int jab_getfd(jconn j)
{
    if(j)
        return j->fd;
    else
        return -1;
}

/*
 *  jab_getjid -- get jid structure of user
 *
 *  parameters
 *      j -- connection
 */
jid jab_getjid(jconn j)
{
    if(j)
        return(j->user);
    else
        return NULL;
}

/*  jab_getsid -- get stream id
 *  This is the id of server's <stream:stream> tag and used for
 *  digest authorization.
 *
 *  parameters
 *      j -- connection
 */
char *jab_getsid(jconn j)
{
    if(j)
        return(j->sid);
    else
        return NULL;
}

/*
 *  jab_getid -- get a unique id
 *
 *  parameters
 *      j -- connection
 */
char *jab_getid(jconn j)
{
    snprintf(j->idbuf, 8, "%d", j->id++);
    return &j->idbuf[0];
}

/*
 *  jab_send -- send xml data
 *
 *  parameters
 *      j -- connection
 *      x -- xmlnode structure
 */
/* old not syncronised funktion
void jab_send(jconn j, xmlnode x)
{
    if (j && j->state != JCONN_STATE_OFF)
    {
            char *buf = xmlnode2str(x);
	    if (buf) pth_write(j->fd, buf, strlen(buf));
#ifdef JDEBUG
	    printf ("out: %s\n", buf);
#endif
    }
}
*/

void jab_send(jconn j, xmlnode x)
{ 
   if(j && x)
     {
       
       Ns_MutexLock(j->wlock); //lock our j->fd for writing 
      
       while(1)
	 { 
           if (j->state == JCONN_STATE_OFF){ break;} //may dump not send data 
            
	    FD_SET(j->fd , j->fdsetw);
            
      
             if( select(j->fd+1 ,NULL , j->fdsetw , NULL , j->tv) > 0 && FD_ISSET(j->fd,j->fdsetw))
	       {   char *buf = xmlnode2str(x);
                   FD_CLR(j->fd,j->fdsetw);
                   if(buf)write(j->fd, buf, strlen(buf));
                   break;
               }
             else 
               { 
	         Ns_RWLockUnlock(j->lock);   // Releas the read lock so other 
                 Ns_ThreadYield();            // threads can change j->state 
                 Ns_RWLockRdLock(j->lock);   // while we are sleeping
               } 
      
             
         }

       Ns_MutexUnlock(j->wlock);
       //xmlnode_free(x);  
     }
  

}
/*
 *  jab_send_raw -- send a string
 *
 *  parameters
 *      j -- connection
 *      str -- xml string
 */
void jab_send_raw(jconn j, const char *str)
{  if(j && str)
     {
      
       Ns_MutexLock(j->wlock); //Lock for writing to j->fd
      
       while(1)
	 { 
           if (j->state == JCONN_STATE_OFF)break; //may dump not send data 
            
	     
	         FD_SET(j->fd , j->fdsetw);
           
      
             if( select(j->fd+1 ,NULL , j->fdsetw , NULL , j->tv) > 0 && FD_ISSET(j->fd,j->fdsetw))
	       {   FD_CLR(j->fd,j->fdsetw);
                   write(j->fd, str, strlen(str));
                   break;
               }
             else 
               {
                 Ns_RWLockUnlock(j->lock);   // Releas the read lock so other 
                 Ns_ThreadYield();            // threads can change j->state 
                 Ns_RWLockRdLock(j->lock);   // while we are sleeping         
               } 
      
             
         }

       Ns_MutexUnlock(j->wlock); //releas Lock for writing to j->fd
       
     }
#ifdef JDEBUG
    printf ("out: %s\n", str);
#endif
}

/*
 *  jab_recv -- read and parse incoming data
 *
 *  parameters
 *      j -- connection
 */
void jab_recv(jconn j , int ssize)
{
  //static char StandartBuff[SockBuffStandartSize];
    static int lastsize = 0; 
    int len,rsize;
    static char buf[16000];
    //char *buf = NULL;
    //pool p = NULL;

    //p = pool_heap(ssize + 1024); 
    //buf = pmalloc(p , sizeof(char) * ssize);
    
    /*if(SockBuffStandartSize > ssize && rsize > ssize) {buf = &SandartBuff;}
      else {p = pool_heap(ssize + 100); buf = pmalloc(p , sizeof(char) * ssize); rzise = ssize;}*/


    if(!j || j->state == JCONN_STATE_OFF)
      {//if(p) pool_free(p); 
       return;
      }

    len = read(j->fd, buf, 16000);
    /* Ns_Log added 
    Ns_Log(Notice,"jab_receive");	
    Ns_Log(Notice,"buffer :%s ", buf);
    Ns_Log(Notice,"size :%i ", len);*/
    if(len>0)
    {   
        buf[len] = '\0';
	Ns_Log(Notice,"Buffer inhalt von jab_recive :%s ", buf);
#ifdef JDEBUG
        printf (" in: %s\n", buf);
#endif
        XML_Parse(j->parser, buf, len, 0);
    }
    /* < changed to <= (if select returns 1, but size of read =0 server probably down(?) */
    else if(len<=0)
    {
       //STATE_EVT(JCONN_STATE_OFF);
        jab_stop(j);
    }

    //if(p) pool_free(p);
}

/*
 *  jab_poll -- check socket for incoming data
 *
 *  parameters
 *      j -- connection
 *      timeout -- poll timeout
 */
void jab_poll(jconn j, int timeout)
{int ssize = 0;
    fd_set fds;
    struct timeval tv;

    if (!j || j->state == JCONN_STATE_OFF)
        return;
       
    FD_ZERO(&fds);
    FD_SET(j->fd, &fds);
   
    if (timeout < 0)
    {
        if ((ssize = select(j->fd + 1, &fds, NULL, NULL, NULL)) > 0)
            jab_recv(j , ssize);
    }
    else
    {
	    tv.tv_sec  = 0;
	    tv.tv_usec = timeout;
	    if ((ssize = select(j->fd + 1, &fds, NULL, NULL, &tv)) > 0)
	      {jab_recv(j , ssize);}
            else
              {} 
    }
}

/*
 *  jab_auth -- authorize user
 *
 *  parameters
 *      j -- connection
 *
 *  returns
 *      id of the iq packet
 */
char *jab_auth(jconn j)
{
    xmlnode x,y,z;
    char *hash, *user, *id;

    if(!j) return(NULL);

    x = jutil_iqnew(JPACKET__SET, NS_AUTH);
    id = jab_getid(j);
    xmlnode_put_attrib(x, "id", "290auth");
    y = xmlnode_get_tag(x,"query");

    user = j->user->user;

    if (user)
    {
        z = xmlnode_insert_tag(y, "username");
        xmlnode_insert_cdata(z, user, -1);
    }

    z = xmlnode_insert_tag(y, "resource");
    xmlnode_insert_cdata(z, j->user->resource, -1);

    if (j->sid)
    {
        z = xmlnode_insert_tag(y, "digest");
        hash = pmalloc(x->p, strlen(j->sid)+strlen(j->pass)+1);
        strcpy(hash, j->sid);
        strcat(hash, j->pass);
        hash = shahash(hash);
        xmlnode_insert_cdata(z, hash, 40);
    }
    else
    {
	z = xmlnode_insert_tag(y, "password");
	xmlnode_insert_cdata(z, j->pass, -1);
    }

    jab_send(j, x);
    xmlnode_free(x);
    return id;
}

/*
 *  jab_reg -- register user
 *
 *  parameters
 *      j -- connection
 *
 *  returns
 *      id of the iq packet
 */
char *jab_reg(jconn j)
{
    xmlnode x,y,z;
    char *hash, *user, *id;

    if (!j) return(NULL);

    x = jutil_iqnew(JPACKET__SET, NS_REGISTER);
    id = jab_getid(j);
    xmlnode_put_attrib(x, "id", id);
    y = xmlnode_get_tag(x,"query");

    user = j->user->user;

    if (user)
    {
        z = xmlnode_insert_tag(y, "username");
        xmlnode_insert_cdata(z, user, -1);
    }

    z = xmlnode_insert_tag(y, "resource");
    xmlnode_insert_cdata(z, j->user->resource, -1);

    if (j->pass)
    {
	z = xmlnode_insert_tag(y, "password");
	xmlnode_insert_cdata(z, j->pass, -1);
    }

    jab_send(j, x);
    xmlnode_free(x);
    j->state = JCONN_STATE_ON;
    STATE_EVT(JCONN_STATE_ON)
    return id;
}


/* local functions */

static void startElement(void *userdata, const char *name, const char **attribs)
{
    xmlnode x;
    jconn j = (jconn)userdata;

    if(j->current)
    {
        /* Append the node to the current one */
        x = xmlnode_insert_tag(j->current, name);
        xmlnode_put_expat_attribs(x, attribs);

        j->current = x;
    }
    else
    {
        x = xmlnode_new_tag(name);
        xmlnode_put_expat_attribs(x, attribs);
        if(strcmp(name, "stream:stream") == 0) {
            /* special case: name == stream:stream */
            /* id attrib of stream is stored for digest auth */
            /* WE DON'T WANNA USE DIGEST AUTH (FOR NOW :) )*/
	    //j->sid = xmlnode_get_attrib(x, "id");
            /* STATE_EVT(JCONN_STATE_AUTH) */
        } else {
            j->current = x;
        }
    }
}

static void endElement(void *userdata, const char *name)
{
    jconn j = (jconn)userdata;
    xmlnode x;
    jpacket p;

    if(j->current == NULL) {
        /* we got </stream:stream> */
        STATE_EVT(JCONN_STATE_OFF)
        return;
    }

    x = xmlnode_get_parent(j->current);

    if(x == NULL)
    {
        /* it is time to fire the event */
        p = jpacket_new(j->current);

        if(j->on_packet)
            (j->on_packet)(j, p);
        else
            xmlnode_free(j->current);
    }

    j->current = x;
}

static void charData(void *userdata, const char *s, int slen)
{
    jconn j = (jconn)userdata;

    if (j->current)
        xmlnode_insert_cdata(j->current, s, slen);
}


















